@@ -1,8 +1,12 @@
 from .ninjax import run
+from .ninjax import reset
 from .ninjax import state
 from .ninjax import rng
 from .ninjax import scope
 from .ninjax import Module
+from .ninjax import Variable
+from .ninjax import cond
+from .ninjax import scan
 from .ninjax import grad
 from .ninjax import HaikuModule
 from .ninjax import FlaxModule
