@@ -4,23 +4,43 @@ import inspect
 import re
 
 import jax
+import jax.numpy as jnp
 
 
 STATE = [None]
 SCOPE = ['']
 RNG = [None]
+RESERVE = [None]
 
 
-def run(fn, state, rng, *a, **k):
+def reset():
+  # global STATE, SCOPE, RNG, RESERVE
+  ModuleMeta.COUNTERS.clear()
+  # STATE[0] = None
+  # SCOPE[0] = ''
+  # RNG[0] = None
+  # RESERVE[0] = []
+
+
+def run(fn, state, rng, *args, nested=False, **kwargs):
   """Run a function or method that uses the global state or the global RNG
   key. The new global state and function output are returned."""
   global STATE
-  STATE[0] = state
-  RNG[0] = rng
-  out = fn(*a, **k)
-  STATE[0] = None
-  RNG[0] = None
-  return state, out
+  if not isinstance(state, dict):
+    raise ValueError('Must provide a dict as state.')
+  if (STATE[0] is not None) and (not nested):
+    print(STATE)
+    raise RuntimeError(
+        'If you really want to nest run() calls, use nested=False.')
+  before = STATE[0], RNG[0], RESERVE[0]
+  try:
+    STATE[0] = state
+    RNG[0] = rng
+    RESERVE[0] = []
+    out = fn(*args, **kwargs)
+  finally:
+    STATE[0], RNG[0], RESERVE[0] = before
+  return out, state
 
 
 def state():
@@ -31,13 +51,21 @@ def state():
   return STATE[0]
 
 
-def rng():
+def rng(amount=None, reserve=16):
   """Split the global RNG key and return a local key."""
-  global RNG
+  global RNG, RESERVE
   if RNG[0] is None:
     raise RuntimeError('Run functions that use the global RNG key with run().')
-  RNG[0], rng = jax.random.split(RNG[0])
-  return rng
+  if amount:
+    keys = jax.random.split(RNG[0], amount + 1)
+    RNG[0] = keys[0]
+    return keys[1:]
+  else:
+    if not RESERVE[0]:
+      keys = jax.random.split(RNG[0], reserve)
+      RNG[0] = keys[0]
+      RESERVE[0] = list(keys[1:])
+    return RESERVE[0].pop(0)
 
 
 @contextlib.contextmanager
@@ -90,14 +118,15 @@ class ModuleMeta(type):
     name and path for the module instance."""
     obj = cls.__new__(cls)
     name = name or cls.__name__
-    if name in cls.COUNTERS:
-      cls.COUNTERS[name] += 1
-      name += str(cls.COUNTERS[name])
+    global SCOPE
+    path = SCOPE[0] + '/' + name
+    if path in cls.COUNTERS:
+      cls.COUNTERS[path] += 1
+      path += str(cls.COUNTERS[path])
     else:
-      cls.COUNTERS[name] = 1
-    obj.name = name
-    with scope(name) as path:
-      obj._path = path
+      cls.COUNTERS[path] = 1
+    obj._path = path
+    obj._submodules = {}
     init = _scope_method(cls.__init__)
     init(obj, *args, **kwargs)
     return obj
@@ -125,31 +154,43 @@ class Module(object, metaclass=ModuleMeta):
     return self._path
 
   def get(self, name, *args, **kwargs):
+    # TODO: Add separate functions for accessing modules and numeric state.
     """Retrieve or create a state entry that belongs to this module."""
     state_ = state()
-    name = self.path + '/' + name
-    if name not in state_:
-      ctor, *args = args
-      state_[name] = ctor(*args, **kwargs)
-    return state_[name]
+    path = self.path + '/' + name
+    if name in self._submodules:
+      return self._submodules[name]
+    if path in state_:
+      return state_[path]
+    ctor, *args = args
+    if 'name' in inspect.signature(ctor).parameters:
+      kwargs['name'] = name
+    value = ctor(*args, **kwargs)
+    flat, _ = jax.tree_util.tree_flatten(value)
+    if all(isinstance(x, jnp.ndarray) for x in flat):
+      state_[path] = value
+    else:
+      self._submodules[name] = value
+    return value
 
   def put(self, name, value):
     """Update or create a single state entry that belongs to this module."""
     self.set_state({self.path + '/' + name: value})
+    return value
 
-  def get_state(self, filter=r'.*', allow_empty=False):
+  def get_state(self, pattern=r'.*', allow_empty=False):
     """Read the state entries of this module, optionally filtered by regex."""
     state_ = state()
-    filter = re.compile(filter)
+    pattern = re.compile(pattern)
     prefix = self.path + '/'
     results = {}
     for key, value in state_.items():
       if not key.startswith(prefix):
         continue
-      if filter.match(key[len(prefix):]):
+      if pattern.match(key[len(prefix):]):
         results[key] = value
     if not allow_empty and not results:
-      raise KeyError(f'Filter {filter} matched no state keys.')
+      raise KeyError(f'Pattern {pattern} matched no state keys.')
     return results
 
   def set_state(self, mapping):
@@ -161,21 +202,76 @@ class Module(object, metaclass=ModuleMeta):
     state().update(mapping)
 
 
-def grad(fn, keys):
+class Variable(Module):
+
+  def __init__(self, ctor, *args, **kwargs):
+    self.ctor = ctor
+    self.args = args
+    self.kwargs = kwargs
+
+  def read(self):
+    return self.get('value', self.ctor, *self.args, **self.kwargs)
+
+  def write(self, value):
+    return self.put('value', value)
+
+
+def grad(fn, keys, has_aux=False):
   """Compute the value and gradient of a function with respect to the state
   entries of the provided keys."""
   state_ = state()
+  before = state_.copy()
   def inner(x, *args, **kwargs):
     state_.update(x)
-    return fn(*args, **kwargs)
-  grad_fn = jax.value_and_grad(inner)
+    out = fn(*args, **kwargs)
+    return out
+  grad_fn = jax.value_and_grad(inner, has_aux=has_aux)
   @functools.wraps(grad_fn)
   def wrapper(*args, **kwargs):
     x = {k: state_[k] for k in keys}
-    return grad_fn(x, *args, **kwargs)
+    out = grad_fn(x, *args, **kwargs)
+    state_.update(before)
+    return out
   return wrapper
 
 
+def cond(pred, true_fn, false_fn, *operands):
+  global STATE
+  out, state = jax.lax.cond(
+      pred,
+      lambda rng1, rng2, *args: run(true_fn, STATE[0], rng1, *args, nested=True),
+      lambda rng1, rng2, *args: run(false_fn, STATE[0], rng2, *args, nested=True),
+      *rng(2), *operands)
+  STATE[0] = state
+  return out
+
+
+def scan(f, init, xs, length=None, reverse=False, unroll=1):
+  global STATE
+
+  # # Run once outside of LAX to initialize variables, but do not update
+  # # variables that already exist to reduce side effects.
+  # x = jax.tree_util.tree_map(lambda x: x[0], xs)
+  # state = run(f, STATE[0].copy(), rng, init, x, nested=True)[1]
+  # print('A', STATE[0])
+  # for key, value in state.items():
+  #   if key not in STATE[0]:
+  #     STATE[0][key] = value
+  # print('B', STATE[0])
+
+  def inner(carry, x):
+    carry, state = carry
+    x, rng = x
+    (carry, y), state = run(f, state, rng, carry, x, nested=True)
+    return (carry, state), y
+  rngs = rng(length or len(jax.tree_util.tree_flatten(xs)[0][0]))
+  carry, ys = jax.lax.scan(
+      inner, (init, STATE[0]), (xs, rngs), length, reverse, unroll)
+  carry, state = carry
+  STATE[0] = state
+  return carry, ys
+
+
 class HaikuModule(Module):
 
   def __init__(self, ctor, *args, **kwargs):
@@ -185,7 +281,7 @@ class HaikuModule(Module):
     self.transformed = hk.transform(net)
 
   def __call__(self, *args, **kwargs):
-    state = self.get('haiku', self.transformed.init, rng(), *args, **kwargs)
+    state = self.get('state', self.transformed.init, rng(), *args, **kwargs)
     return self.transformed.apply(state, rng(), *args, **kwargs)
 
 
@@ -195,7 +291,7 @@ class FlaxModule(Module):
     self.module = ctor(*args, **kwargs)
 
   def __call__(self, *args, **kwargs):
-    state = self.get('flax', self.module.init, rng(), *args, **kwargs)
+    state = self.get('state', self.module.init, rng(), *args, **kwargs)
     return self.module.apply(state, *args, **kwargs)
 
 
